# -*- coding: utf-8 -*-
#
# Copyright 2018 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Client for handling a local repository."""

import datetime
import uuid
from contextlib import contextmanager
from subprocess import PIPE, STDOUT, call

import attr
import filelock
import yaml
from werkzeug.utils import cached_property, secure_filename

from renku._compat import Path

HAS_LFS = call(['git', 'lfs'], stdout=PIPE, stderr=STDOUT) == 0


@attr.s
class PathMixin(object):
    """Define a default path attribute."""

    path = attr.ib(converter=lambda arg: Path(arg).resolve().absolute())

    @path.default
    def _default_path(self):
        """Return default repository path."""
        from git import InvalidGitRepositoryError
        from renku.cli._git import get_git_home

        try:
            return get_git_home()
        except InvalidGitRepositoryError:
            return '.'

    @path.validator
    def _check_path(self, _, value):
        """Check the path exists and it is a directory."""
        if not (value.exists() and value.is_dir()):
            raise ValueError('Define an existing directory.')


@attr.s
class RepositoryApiMixin(object):
    """Client for handling a local repository."""

    renku_home = attr.ib(default='.renku')
    """Define a name of the Renku folder (default: ``.renku``)."""

    renku_path = attr.ib(init=False)
    """Store a ``Path`` instance of the Renku folder."""

    git = attr.ib(init=False)
    """Store an instance of the Git repository."""

    parent = attr.ib(default=None)
    """Store a pointer to the parent repository."""

    METADATA = 'metadata.yml'
    """Default name of Renku config file."""

    LOCK_SUFFIX = '.lock'
    """Default suffix for Renku lock file."""

    WORKFLOW = 'workflow'
    """Directory for storing workflow in Renku."""

    def __attrs_post_init__(self):
        """Initialize computed attributes."""
        from git import InvalidGitRepositoryError, Repo

        #: Configure Renku path.
        path = Path(self.renku_home)
        if not path.is_absolute():
            path = self.path / path

        path.relative_to(path)
        self.renku_path = path

        #: Create an instance of a Git repository for the given path.
        try:
            self.git = Repo(str(self.path))
        except InvalidGitRepositoryError:
            self.git = None
        # TODO except

    @property
    def lock(self):
        """Create a Renku config lock."""
        return filelock.FileLock(
            str(self.renku_path.with_suffix(self.LOCK_SUFFIX))
        )

    @property
    def renku_metadata_path(self):
        """Return a ``Path`` instance of Renku metadata file."""
        return self.renku_path.joinpath(self.METADATA)

    @property
    def workflow_path(self):
        """Return a ``Path`` instance of the workflow folder."""
        return self.renku_path / self.WORKFLOW

    @cached_property
    def cwl_prefix(self):
        """Return a CWL prefix."""
        self.workflow_path.mkdir(parents=True, exist_ok=True)  # for Python 3.5
        return str(self.workflow_path.resolve().relative_to(self.path))

    def is_cwl(self, path):
        """Check if the path is a valid CWL file."""
        return path.startswith(self.cwl_prefix) and path.endswith('.cwl')

    def find_previous_commit(self, paths, revision='HEAD'):
        """Return a previous commit for a given path."""
        file_commits = list(self.git.iter_commits(revision, paths=paths))

        if not file_commits:
            raise KeyError(
                'Could not find a file {0} in range {1}'.format(
                    paths, revision
                )
            )

        return file_commits[0]

    @contextmanager
    def with_metadata(self):
        """Yield an editable metadata object."""
        with self.lock:
            from renku.models._jsonld import asjsonld
            from renku.models.projects import Project

            metadata_path = self.renku_metadata_path

            if self.renku_metadata_path.exists():
                with metadata_path.open('r') as f:
                    source = yaml.load(f) or {}
                metadata = Project.from_jsonld(source)
            else:
                source = {}
                metadata = Project()

            yield metadata

            source.update(**asjsonld(metadata))
            with metadata_path.open('w') as f:
                yaml.dump(source, f, default_flow_style=False)

    @contextmanager
    def with_workflow_storage(self):
        """Yield a workflow storage."""
        with self.lock:
            from renku.models.cwl._ascwl import ascwl
            from renku.models.cwl.workflow import Workflow

            workflow = Workflow()
            yield workflow

            for step in workflow.steps:
                step_name = '{0}_{1}.cwl'.format(
                    uuid.uuid4().hex,
                    secure_filename('_'.join(step.run.baseCommand)),
                )

                workflow_path = self.workflow_path
                if not workflow_path.exists():
                    workflow_path.mkdir()

                step_path = workflow_path / step_name
                with step_path.open('w') as step_file:
                    yaml.dump(
                        ascwl(
                            # filter=lambda _, x: not (x is False or bool(x)
                            step.run,
                            filter=lambda _, x: x is not None,
                            basedir=workflow_path,
                        ),
                        stream=step_file,
                        default_flow_style=False
                    )

    def init_repository(
        self, name=None, force=False, use_external_storage=True
    ):
        """Initialize a local Renku repository."""
        from git import Repo

        path = self.path.absolute()
        if force:
            self.renku_path.mkdir(parents=True, exist_ok=force)
            if self.git is None:
                self.git = Repo.init(str(path))
        else:
            if self.git is not None:
                raise FileExistsError(self.git.git_dir)

            self.renku_path.mkdir(parents=True, exist_ok=force)
            self.git = Repo.init(str(path))

        self.git.description = name or path.name

        # Check that an author can be determined from Git.
        from renku.models.datasets import Author
        Author.from_git(self.git)

        # TODO read existing gitignore and create a unique set of rules
        import pkg_resources
        gitignore_default = pkg_resources.resource_stream(
            'renku.data', 'gitignore.default'
        )
        gitignore_path = path / '.gitignore'
        with gitignore_path.open('w') as gitignore:
            gitignore.write(gitignore_default.read().decode())

            gitignore.write(
                '\n' + str(
                    self.renku_path.relative_to(self.path).
                    with_suffix(self.LOCK_SUFFIX)
                ) + '\n'
            )

        with self.with_metadata() as metadata:
            metadata.name = name
            metadata.updated = datetime.datetime.utcnow()

        # initialize LFS if it is requested and installed
        if use_external_storage and HAS_LFS:
            self.init_external_storage(force=force)

        return str(path)

    def init_external_storage(self, force=False):
        """Initialize the external storage for data."""
        cmd = ['git', 'lfs', 'install', '--local']
        if force:
            cmd.append('--force')

        call(
            cmd,
            stdout=PIPE,
            stderr=STDOUT,
            cwd=str(self.path.absolute()),
        )

    def track_paths_in_storage(self, *paths):
        """Track paths in the external storage."""
        if HAS_LFS and self.git.config_reader(config_level='repository'
                                              ).has_section('filter "lfs"'):
            # FIXME create configurable filter and respect .gitattributes
            paths = [
                path for path in paths if not str(path).endswith('.ipynb')
            ]

            call(
                ['git', 'lfs', 'track'] + list(paths),
                stdout=PIPE,
                stderr=STDOUT,
                cwd=str(self.path),
            )



# -*- coding: utf-8 -*-
#
# Copyright 2017-2018 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility functions for managing the underling Git repository."""

import os
import sys
from contextlib import contextmanager

import click

from renku import errors

GIT_KEY = 'renku.git'


def set_git_home(value):
    """Set Git path."""
    ctx = click.get_current_context()
    ctx.meta[GIT_KEY] = value


def get_git_home():
    """Get Git path from current context."""
    ctx = click.get_current_context(silent=True)
    if ctx and GIT_KEY in ctx.meta:
        return ctx.meta[GIT_KEY]

    from git import Repo
    return Repo('.', search_parent_directories=True).working_dir


def _modified_paths(repo):
    """Return paths of modified files."""
    return [item.b_path for item in repo.index.diff(None) if item.b_path]


def _dirty_paths(repo):
    """Get paths of dirty files in the repository."""
    repo_path = repo.working_dir
    return {
        os.path.join(repo_path, p)
        for p in repo.untracked_files + _modified_paths(repo)
    }


def _mapped_std_streams(lookup_paths, streams=('stdin', 'stdout', 'stderr')):
    """Get a mapping of standard streams to given paths."""
    # FIXME add device number too
    standard_inos = {}
    for stream in streams:
        try:
            stream_stat = os.fstat(getattr(sys, stream).fileno())
            key = stream_stat.st_dev, stream_stat.st_ino
            standard_inos[key] = stream
        except Exception:  # FIXME UnsupportedOperation
            pass
        # FIXME if not getattr(sys, stream).istty()

    def stream_inos(paths):
        """Yield tuples with stats and path."""
        for path in paths:
            try:
                stat = os.stat(path)
                key = (stat.st_dev, stat.st_ino)
                if key in standard_inos:
                    yield standard_inos[key], path
            except FileNotFoundError:  # pragma: no cover
                pass

    return dict(stream_inos(lookup_paths)) if standard_inos else {}


def _clean_streams(repo, mapped_streams):
    """Clean mapped standard streams."""
    for stream_name in ('stdout', 'stderr'):
        stream = mapped_streams.get(stream_name)
        if not stream:
            continue

        path = os.path.relpath(stream, start=repo.working_dir)
        if (path, 0) not in repo.index.entries:
            os.remove(stream)
        else:
            blob = repo.index.entries[(path, 0)].to_blob(repo)
            with open(path, 'wb') as fp:
                fp.write(blob.data_stream.read())


@contextmanager
def with_git(
    clean=True, up_to_date=False, commit=True, ignore_std_streams=False
):
    """Perform Git checks and operations."""
    from git import Repo

    repo_path = get_git_home()
    current_dir = os.getcwd()

    if clean:
        try:
            os.chdir(repo_path)
            repo = Repo(repo_path)

            dirty_paths = _dirty_paths(repo)
            mapped_streams = _mapped_std_streams(dirty_paths)

            if ignore_std_streams:
                if dirty_paths - set(mapped_streams.values()):
                    _clean_streams(repo, mapped_streams)
                    raise errors.DirtyRepository(repo)

            elif repo.is_dirty(untracked_files=True):
                _clean_streams(repo, mapped_streams)
                raise errors.DirtyRepository(repo)
        finally:
            os.chdir(current_dir)

    if up_to_date:
        # TODO
        # Fetch origin/master
        # is_ancestor('origin/master', 'HEAD')
        pass

    yield

    if commit:
        try:
            os.chdir(repo_path)
            repo = Repo(get_git_home())
            repo.git.add('--all')
            argv = [os.path.basename(sys.argv[0])] + sys.argv[1:]
            # Ignore pre-commit hooks since we have already done everything.
            repo.index.commit(' '.join(argv), skip_hooks=True)
        finally:
            os.chdir(current_dir)


def _safe_issue_checkout(repo, issue=None):
    """Safely checkout branch for the issue."""
    branch_name = str(issue) if issue else 'master'
    if branch_name not in repo.heads:
        branch = repo.create_head(branch_name)
    else:
        branch = repo.heads[branch_name]
    branch.checkout()



# -*- coding: utf-8 -*-
#
# Copyright 2017, 2018 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""The base command for interacting with the Renku platform.

``renku`` (base command)
------------------------

To list the available commands, either run ``renku`` with no parameters or
execute ``renku help``:

.. code-block:: console

    $ renku help
    Usage: renku [OPTIONS] COMMAND [ARGS]...

    Check common Renku commands used in various situations.

    Options:
      --version            Print version number.
      --config PATH        Location of client config files.
      --config-path        Print application config path.
      --path <path>        Location of a Renku repository.  [default: .]
      --renku-home <path>  Location of Renku directory.  [default: .renku]
      -h, --help           Show this message and exit.

    Commands:
      # [...]

Configuration files
~~~~~~~~~~~~~~~~~~~

Depending on your system, you may find the configuration files used by Renku
command line in a different folder. By default, the following rules are used:

MacOS:
  ``~/Library/Application Support/Renku``
Unix:
  ``~/.config/renku``
Windows:
  ``C:\Users\<user>\AppData\Roaming\Renku``

If in doubt where to look for the configuration file, you can display its path
by running ``renku --config-path``.

You can specify a different location via the ``RENKU_CONFIG`` environment
variable or the ``--config`` command line option. If both are specified, then
the ``--config`` option value is used. For example:

.. code-block:: console

    $ renku --config ~/renku/config/ init

instructs Renku to store the configuration files in your ``~/renku/config/``
directory when running the ``init`` command.
"""

import uuid

import click
import yaml

from ._config import RENKU_HOME, default_config_dir, print_app_config_path
from ._version import print_version
from .config import config
from .dataset import dataset
from .githooks import githooks
from .image import image
from .init import init
from .log import log
from .rerun import rerun
from .run import run
from .runner import runner
from .show import show
from .status import status
from .update import update
from .workflow import workflow
from .workon import deactivate, workon


def _uuid_representer(dumper, data):
    """Add UUID serializer for YAML."""
    return dumper.represent_str(str(data))


yaml.add_representer(uuid.UUID, _uuid_representer)


@click.group(
    context_settings={
        'auto_envvar_prefix': 'RENKU',
        'help_option_names': ['-h', '--help'],
    }
)
@click.option(
    '--version',
    is_flag=True,
    callback=print_version,
    expose_value=False,
    is_eager=True,
    help=print_version.__doc__
)
@click.option(
    '--config',
    envvar='RENKU_CONFIG',
    default=default_config_dir,
    type=click.Path(),
    expose_value=False,
    help='Location of client config files.'
)
@click.option(
    '--config-path',
    is_flag=True,
    callback=print_app_config_path,
    expose_value=False,
    is_eager=True,
    help=print_app_config_path.__doc__
)
@click.option(
    '--path',
    show_default=True,
    metavar='<path>',
    default='.',
    help='Location of a Renku repository.'
)
@click.option(
    '--renku-home',
    envvar='RENKU_HOME',
    show_default=True,
    metavar='<path>',
    default=RENKU_HOME,
    help='Location of Renku directory.'
)
@click.pass_context
def cli(ctx, path, renku_home):
    """Check common Renku commands used in various situations."""


@cli.command()
@click.pass_context
def help(ctx):
    """Show help message and exit."""
    click.echo(ctx.parent.get_help())


# Register subcommands:
cli.add_command(config)
cli.add_command(dataset)
cli.add_command(deactivate)
cli.add_command(githooks)
cli.add_command(image)
cli.add_command(init)
cli.add_command(log)
cli.add_command(rerun)
cli.add_command(run)
cli.add_command(runner)
cli.add_command(status)
cli.add_command(show)
cli.add_command(update)
cli.add_command(workflow)
cli.add_command(workon)



# -*- coding: utf-8 -*-
#
# Copyright 2018 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Update outdated files created by the "run" command.

Recreating outdated files
~~~~~~~~~~~~~~~~~~~~~~~~~

The information about dependencies for each file in the repository is generated
from information stored in the underlying Git repository.

A minimal dependency graph is generated for each outdated file stored in the
repository. It means that only the necessary steps will be executed and the
workflow used to orchestrate these steps is stored in the repository.

Assume that the following history for the file ``H`` exists.

.. code-block:: text

          C---D---E
         /         \
    A---B---F---G---H

The first example shows situation when ``D`` is modified and files ``E`` and
``H`` become outdated.

.. code-block:: text

          C--*D*--(E)
         /          \
    A---B---F---G---(H)

    ** - modified
    () - needs update

In this situation, you can do efectively two things:

* Recreate a single file by running

  .. code-block:: console

     $ renku update E

* Update all files by simply running

  .. code-block:: console

     $ renku update

.. note:: If there were uncommitted changes then the command fails.
   Check :program:`git status` to see details.

Pre-update checks
~~~~~~~~~~~~~~~~~

In the next example, files ``A`` or ``B`` are modified, hence the majority
of dependent files must be recreated.

.. code-block:: text

            (C)--(D)--(E)
           /            \
    *A*--*B*--(F)--(G)--(H)

To avoid excesive recreation of the large portion of files which could have
been affected by a simple change of an input file, consider speficing a single
file (e.g. ``renku update G``). See also :ref:`cli-status`.

.. _cli-update-with-siblings:

Update siblings
~~~~~~~~~~~~~~~

If a tool produces multiple output files, these outputs need to be always
updated together.

.. code-block:: text

                   (B)
                  /
    *A*--[step 1]--(C)
                  \
                   (D)

An attempt to update a single file would fail with the following error.

.. code-block:: console

   $ renku update C
   Error: There are missing output siblings:

        B
        D

   Include the files above in the command or use --with-siblings option.

The following commands will produce the same result.

.. code-block:: console

   $ renku update --with-siblings C
   $ renku update B C D

"""

import sys
import uuid

import click

from renku.models.cwl._ascwl import ascwl

from ._client import pass_local_client
from ._git import with_git
from ._graph import Graph, _safe_path
from ._options import option_siblings


@click.command()
@click.option('--revision', default='HEAD')
@click.option(
    '--no-output',
    is_flag=True,
    default=False,
    help='Display commands without output files.'
)
@option_siblings
@click.argument(
    'paths', type=click.Path(exists=True, dir_okay=False), nargs=-1
)
@pass_local_client
@click.pass_context
@with_git()
def update(ctx, client, revision, no_output, siblings, paths):
    """Update existing files by rerunning their outdated workflow."""
    graph = Graph(client)
    status = graph.build_status(revision=revision, can_be_cwl=no_output)
    paths = {graph.normalize_path(path) for path in paths} \
        if paths else status['outdated'].keys()

    if not paths:
        click.secho(
            'All files were generated from the latest inputs.', fg='green'
        )
        sys.exit(0)

    outputs = graph.build(paths=paths, revision=revision)

    # Check or extend siblings of outputs.
    outputs = siblings(graph, outputs)
    output_paths = {path for _, path in outputs if _safe_path(path)}

    # Get parents of all clean nodes
    import networkx as nx

    clean_paths = set(status['up-to-date'].keys()) - output_paths
    clean_nodes = {(c, p) for (c, p) in graph.G if p in clean_paths}
    clean_parents = set()
    for key in clean_nodes:
        clean_parents |= nx.ancestors(graph.G, key)

    subnodes = set()
    for key in outputs:
        if key in graph.G:
            subnodes |= nx.shortest_path_length(graph.G, target=key).keys()

    graph.G.remove_nodes_from(clean_parents)
    graph.G.remove_nodes_from([n for n in graph.G if n not in subnodes])

    # Store the generated workflow used for updating paths.
    import yaml

    output_file = client.workflow_path / '{0}.cwl'.format(uuid.uuid4().hex)
    with output_file.open('w') as f:
        f.write(
            yaml.dump(
                ascwl(
                    graph.ascwl(global_step_outputs=True),
                    filter=lambda _, x: x is not None,
                    basedir=client.workflow_path,
                ),
                default_flow_style=False
            )
        )

    from ._cwl import execute
    execute(client, output_file, output_paths=output_paths)



# -*- coding: utf-8 -*-
#
# Copyright 2018 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Client for handling datasets."""

import os
import shutil
import stat
import warnings
from contextlib import contextmanager
from urllib import error, parse

import attr
import requests
import yaml

from renku._compat import Path
from renku.models.datasets import Author, Dataset, DatasetFile, NoneType


@attr.s
class DatasetsApiMixin(object):
    """Client for handling datasets."""

    datadir = attr.ib(default='data', converter=str)
    """Define a name of the folder for storing datasets."""

    @contextmanager
    def with_dataset(self, name=None):
        """Yield an editable metadata object for a dataset."""
        with self.lock:
            from renku.models._jsonld import asjsonld
            from renku.models.datasets import Dataset
            path = None
            dataset = None

            dataset_path = self.path / self.datadir / name

            if name:
                path = dataset_path / self.METADATA
                if path.exists():
                    with path.open('r') as f:
                        source = yaml.load(f) or {}
                    dataset = Dataset.from_jsonld(source)

            if dataset is None:
                source = {}
                dataset = Dataset(name=name)
                try:
                    dataset_path.mkdir(parents=True, exist_ok=True)
                except FileExistsError:
                    raise FileExistsError('This dataset already exists.')

            yield dataset

            source.update(
                **asjsonld(
                    dataset,
                    filter=lambda attr, _: attr.name != 'datadir',
                )
            )

            # TODO
            # if path is None:
            #     path = dataset_path / self.METADATA
            #     if path.exists():
            #         raise ValueError('Dataset already exists')

            with path.open('w') as f:
                yaml.dump(source, f, default_flow_style=False)

    def add_data_to_dataset(self, dataset, url, git=False, **kwargs):
        """Import the data into the data directory."""
        dataset_path = self.path / self.datadir / dataset.name
        git = git or check_for_git_repo(url)

        target = kwargs.pop('target', None)

        if git:
            if isinstance(target, (str, NoneType)):
                dataset.files.update(
                    self._add_from_git(
                        dataset, dataset_path, url, target, **kwargs
                    )
                )
            else:
                for t in target:
                    dataset.files.update(
                        self._add_from_git(
                            dataset, dataset_path, url, t, **kwargs
                        )
                    )
        else:
            dataset.files.update(
                self._add_from_url(dataset, dataset_path, url, **kwargs)
            )

    def _add_from_url(self, dataset, path, url, nocopy=False, **kwargs):
        """Process an add from url and return the location on disk."""
        u = parse.urlparse(url)

        if u.scheme not in Dataset.SUPPORTED_SCHEMES:
            raise NotImplementedError(
                '{} URLs are not supported'.format(u.scheme)
            )

        # Respect the directory struture inside the source path.
        relative_to = kwargs.pop('relative_to', None)
        if relative_to:
            dst_path = Path(url).resolve().absolute().relative_to(
                Path(relative_to).resolve().absolute()
            )
        else:
            dst_path = os.path.basename(url)

        dst = path.joinpath(dst_path).absolute()

        if u.scheme in ('', 'file'):
            src = Path(u.path).absolute()

            # if we have a directory, recurse
            if src.is_dir():
                files = {}
                dst.mkdir(parents=True, exist_ok=True)
                for f in src.iterdir():
                    files.update(
                        self._add_from_url(
                            dataset,
                            dst,
                            f.absolute().as_posix(),
                            nocopy=nocopy
                        )
                    )
                return files

            # Make sure the parent directory exists.
            dst.parent.mkdir(parents=True, exist_ok=True)

            if nocopy:
                try:
                    os.link(str(src), str(dst))
                except Exception as e:
                    raise Exception(
                        'Could not create hard link '
                        '- retry without nocopy.'
                    ) from e
            else:
                shutil.copy(str(src), str(dst))

            # Do not expose local paths.
            src = None
        else:
            try:
                response = requests.get(url)
                dst.write_bytes(response.content)
            except error.HTTPError as e:  # pragma nocover
                raise e

        # make the added file read-only
        mode = dst.stat().st_mode & 0o777
        dst.chmod(mode & ~(stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH))

        self.track_paths_in_storage(str(dst.relative_to(self.path)))
        dataset_path = self.path / self.datadir / dataset.name
        result = dst.relative_to(dataset_path).as_posix()
        return {
            result:
                DatasetFile(
                    path=result,
                    url=url,
                    authors=dataset.authors,
                    dataset=dataset.name,
                )
        }

    def _add_from_git(self, dataset, path, url, target, **kwargs):
        """Process adding resources from another git repository.

        The submodules are placed in ``.renku/vendors`` and linked
        to the *path* specified by the user.
        """
        from git import Repo

        # create the submodule
        u = parse.urlparse(url)
        submodule_path = self.renku_path / 'vendors' / (u.netloc or 'local')

        # Respect the directory struture inside the source path.
        relative_to = kwargs.get('relative_to', None)

        if u.scheme in ('', 'file'):
            warnings.warn('Importing local git repository, use HTTPS')
            # determine where is the base repo path
            r = Repo(url, search_parent_directories=True)
            src_repo_path = Path(r.git_dir).parent
            submodule_name = src_repo_path.name
            submodule_path = submodule_path / str(src_repo_path).lstrip('/')

            # if repo path is a parent, rebase the paths and update url
            if src_repo_path != Path(u.path):
                top_target = Path(
                    u.path
                ).resolve().absolute().relative_to(src_repo_path)
                if target:
                    target = top_target / target
                else:
                    target = top_target
                url = src_repo_path.as_posix()
        elif u.scheme in ('http', 'https'):
            submodule_name = os.path.splitext(os.path.basename(u.path))[0]
            submodule_path = submodule_path.joinpath(
                os.path.dirname(u.path).lstrip('/'), submodule_name
            )
        else:
            raise NotImplementedError(
                'Scheme {} not supported'.format(u.scheme)
            )

        # FIXME: do a proper check that the repos are not the same
        if submodule_name not in (s.name for s in self.git.submodules):
            # new submodule to add
            self.git.create_submodule(
                name=submodule_name, path=submodule_path.as_posix(), url=url
            )

        src = submodule_path / (target or '')

        if target and relative_to:
            relative_to = Path(relative_to)
            if relative_to.is_absolute():
                assert u.scheme in {
                    '', 'file'
                }, ('Only relative paths can be used with URLs.')
                target = (Path(url).resolve().absolute() / target).relative_to(
                    relative_to.resolve()
                )
            else:
                # src already includes target so we do not have to append it
                target = src.relative_to(submodule_path / relative_to)

        # link the target into the data directory
        dst = self.path / path / submodule_name / (target or '')

        # if we have a directory, recurse
        if src.is_dir():
            files = {}
            dst.mkdir(parents=True, exist_ok=True)
            # FIXME get all files from submodule index
            for f in src.iterdir():
                try:
                    files.update(
                        self._add_from_git(
                            dataset,
                            path,
                            url,
                            target=f.relative_to(submodule_path),
                            **kwargs
                        )
                    )
                except ValueError:
                    pass  # skip files outside the relative path
            return files

        if not dst.parent.exists():
            dst.parent.mkdir(parents=True)

        os.symlink(os.path.relpath(str(src), str(dst.parent)), str(dst))

        # grab all the authors from the commit history
        git_repo = Repo(str(submodule_path.absolute()))
        authors = []
        for commit in git_repo.iter_commits(paths=target):
            author = Author.from_commit(commit)
            if author not in authors:
                authors.append(author)

        dataset_path = self.path / self.datadir / dataset.name
        result = dst.relative_to(dataset_path).as_posix()

        if u.scheme in ('', 'file'):
            url = None
        else:
            url = '{}/{}'.format(url, target)

        return {
            result:
                DatasetFile(
                    path=result,
                    url=url,
                    authors=authors,
                    dataset=dataset.name,  # TODO detect original dataset
                )
        }


def check_for_git_repo(url):
    """Check if a url points to a git repository."""
    u = parse.urlparse(url)
    is_git = False

    if os.path.splitext(u.path)[1] == '.git':
        is_git = True
    elif u.scheme in ('', 'file'):
        from git import InvalidGitRepositoryError, Repo

        try:
            Repo(u.path, search_parent_directories=True)
            is_git = True
        except InvalidGitRepositoryError:
            is_git = False
    return is_git



# -*- coding: utf-8 -*-
#
# Copyright 2018 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Show information about objects in current repository.

Siblings
~~~~~~~~

In situations when multiple outputs have been generated by a single
``renku run`` command, the siblings can be discovered by running
``renku show siblings PATH`` command.

Assume that the following graph represents relations in the repository.

.. code-block:: text

          D---E---G
         /     \
    A---B---C   F

Then the following outputs would be shown.

.. code-block:: console

   $ renku show siblings C
   C
   D
   $ renku show siblings G
   F
   G
   $ renku show siblings A
   A


Output files
~~~~~~~~~~~~

You can list all output files generated in the repository by running
``renku show outputs`` commands. Alternatively, you can check if all
paths specified as arguments are output files.

.. code-block:: console

   $ renku run wc < source.txt > result.wc
   $ renku show outputs
   result.wc
   $ renku show outputs source.txt
   $ $?  # last command finished with an error code
   1

"""

import click

from renku._compat import Path

from ._client import pass_local_client
from ._git import with_git
from ._graph import Graph


@click.group()
def show():
    """Show information about objects in current repository.

    NOTE: The command produces machine readable output.
    """


@show.command()
@click.option('--revision', default='HEAD')
@click.argument(
    'paths', type=click.Path(exists=True, dir_okay=False), nargs=-1
)
@pass_local_client
@with_git(clean=False, commit=False)
def siblings(client, revision, paths):
    """Show siblings for given paths."""
    graph = Graph(client)
    paths = [graph.normalize_path(path) for path in paths]
    nodes = graph.build(paths=paths, revision=revision)
    siblings_ = set(nodes)
    for node in nodes:
        siblings_ |= graph.siblings(node)

    for _, path in siblings_:
        click.echo(graph._format_path(path))


@show.command()
@click.option('--revision', default='HEAD')
@click.argument(
    'paths',
    type=click.Path(exists=True, dir_okay=False),
    nargs=-1,
)
@pass_local_client
@click.pass_context
@with_git(clean=False, commit=False)
def outputs(ctx, client, revision, paths):
    r"""Show output files in the repository.

    <PATHS>    Files to show. If no files are given all output files are shown.
    """
    graph = Graph(client)
    paths = [graph.normalize_path(path) for path in paths]
    filter = None

    if not paths:
        graph.build_status(revision=revision)
    else:
        filter = set(graph.build(paths=paths, revision=revision))

    output_paths = {
        Path(output_key[1]).resolve()
        for (tool_key, output_key) in graph.G.edges
        if 'tool' in graph.G.nodes[tool_key] and
        (filter is None or output_key in filter)
    }
    click.echo('\n'.join(graph._format_path(path) for path in output_paths))
    ctx.exit(0 if filter is None or len(output_paths) == len(filter) else 1)



# -*- coding: utf-8 -*-
#
# Copyright 2018 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Graph builder."""

import os
from collections import deque

import attr
import yaml

from renku.api import LocalClient
from renku.models._datastructures import DirectoryTree
from renku.models.cwl._ascwl import CWLClass


@attr.s
class Dependency(object):
    """Represent a dependent path."""

    commit = attr.ib()
    client = attr.ib()

    path = attr.ib(default=None)
    id = attr.ib(default=None)

    submodules = attr.ib(default=attr.Factory(list))

    @classmethod
    def from_revision(cls, client, path, revision='HEAD', **kwargs):
        """Return dependency from given path and revision."""
        return cls(
            client=client,
            commit=client.find_previous_commit(path, revision=revision),
            path=path,
            **kwargs
        )


@attr.s
class Action(object):
    """Represent an action in the repository."""

    commit = attr.ib()
    client = attr.ib()

    process = attr.ib(default=None)
    process_path = attr.ib(default=None)
    inputs = attr.ib(default=attr.Factory(dict))
    outputs = attr.ib(default=attr.Factory(dict))

    parent = attr.ib(default=None)
    children = attr.ib(default=attr.Factory(list))

    submodules = attr.ib(default=attr.Factory(list))

    def change_key(self, dependency, path=None):
        """Rename the path part in the key."""
        path = path or dependency.path
        return str(dependency.commit), path

    def iter_nodes(self, expand_workflow=True):
        """Yield all graph nodes."""
        for path, _ in self.outputs.items():
            data = {
                'commit': self.commit,
                'path': path,
                'submodule': self.submodules,
                'client': self.client,
            }
            yield (str(self.commit), path), data

        if self.process_path:
            if expand_workflow and self.children:
                steps = {step.id: step for step in self.process.steps}
                basedir = os.path.dirname(self.process_path)
                default_data = {
                    'client': self.client,
                    'workflow': self.process,
                    'submodule': self.submodules,
                }
                revision = '{0}^'.format(self.commit)

                for step_id, subprocess in self.children.items():
                    step = steps[step_id]
                    path = os.path.join(basedir, step.run)
                    # The workflow path must be relative to the current tool
                    # since we might be inside a submodule.
                    workflow_path = os.path.relpath(
                        self.process_path, start=os.path.dirname(path)
                    )

                    data = {
                        'path': path,
                        'commit':
                            self.client.find_previous_commit(
                                path, revision=revision
                            ),
                        'tool': subprocess,
                        'workflow_path':
                            '{workflow_path}#steps/{step.id}'.format(
                                workflow_path=workflow_path, step=step
                            ),
                    }
                    data.update(**default_data)
                    yield (str(self.commit), path), data
            else:
                data = {
                    'client': self.client,
                    'commit': self.commit,
                    'path': self.process_path,
                    'tool': self.process,
                    'submodule': self.submodules,
                }
                # TODO submodule
                yield (str(self.commit), self.process_path), data

        for path, dependency in self.inputs.items():
            data = {
                'commit': dependency.commit,
                'path': dependency.path,
                'submodule': dependency.submodules,
                'client': dependency.client,
            }
            yield self.change_key(dependency, path=path), data

    def iter_edges(self, expand_workflow=True):
        """Yield all graph edges."""
        if expand_workflow and self.children:
            commit = self.commit
            workflow = self.process
            path = self.process_path
            basedir = os.path.dirname(self.process_path)

            # Keep track of node identifiers for steps, inputs and outputs:
            step_map = {}
            input_map = {
                dep.id: (str(dep.commit), path)
                for path, dep in self.inputs.items()
            }
            output_map = {}

            for step in workflow.steps:
                step_tool = self.children[step.id]
                tool_path = os.path.join(basedir, step.run)
                tool_key = (str(commit), tool_path)
                step_map[step.id] = tool_key

                for input_id, input_path in step_tool.iter_input_files(
                    basedir
                ):
                    if input_path in commit.stats.files:
                        #: Check intermediate committed files
                        input_key = (str(commit), input_path)
                        output_map[step.id + '/' + input_id] = input_key
                        #: Edge from an input to the tool.
                        yield input_key, tool_key, {'id': input_id}
                    else:
                        #: Global workflow input
                        source = step.in_[input_id]
                        yield input_map[source], tool_key, {'id': input_id}

                # Find ALL siblings that MUST be generated in the same commit.
                step_outputs = workflow._step_outputs.get(step.id)
                if step_outputs is not None:
                    step_outputs = step_outputs.items()
                else:
                    step_outputs = step_tool.iter_output_files(
                        basedir, commit=commit
                    )

                for output_id, output_path in step_outputs:
                    node_key = (str(commit), output_path)
                    yield tool_key, node_key, {'id': output_id}

            for step in workflow.steps:
                for alias, source in step.in_.items():
                    name = step.id + '/' + alias

                    if name in output_map and '/' in source:
                        other_step, id_ = source.split('/')
                        other_key = step_map[other_step]
                        yield other_key, output_map[name], {'id': id_}

        elif self.process:
            tool_key = (str(self.commit), self.process_path)
            for path, dependency in self.inputs.items():
                input_key = self.change_key(dependency, path=path)
                #: Edge from an input to the tool.
                yield input_key, tool_key, {'id': dependency.id}

            for path, output_id in self.outputs.items():
                node_key = (str(self.commit), path)
                #: Edge from the tool to an output.
                yield tool_key, node_key, {'id': output_id}

    @classmethod
    def from_git_commit(cls, commit, client, submodules=None):
        """Populate information from the given Git commit."""
        process = None
        process_path = None
        inputs = {}
        outputs = {}
        children = {}
        hierarchy = list(submodules) or []

        tree = DirectoryTree()

        for file_ in commit.stats.files.keys():
            # 1.a Find process (CommandLineTool or Workflow);
            if client.is_cwl(file_):
                if process_path is not None:
                    raise ValueError(file_)  # duplicate
                process_path = file_
                continue

            # Build tree index.
            tree.add(file_)

        if process_path:
            basedir = os.path.dirname(process_path)
            try:
                data = (commit.tree / process_path).data_stream.read()
                process = CWLClass.from_cwl(yaml.load(data))
            except KeyError:
                pass
        else:
            outputs = {path: None for path in tree}

        # 2. Map all outputs;
        if process:
            for output_id, output_path in process.iter_output_files(
                basedir, commit=commit
            ):
                outputs[output_path] = output_id

                # Expand directory entries.
                for subpath in tree.get(output_path, []):
                    outputs.setdefault(
                        os.path.join(output_path, subpath), output_id
                    )

            if process.__class__.__name__ == 'Workflow':
                children = process._tools

        # 3. Identify input files (filepath: (input_id, commit))
        if process and process_path:
            revision = '{0}^'.format(commit)

            try:
                from git import Submodule

                submodules = [
                    submodule for submodule in Submodule.
                    iter_items(client.git, parent_commit=commit)
                ]
            except (RuntimeError, ValueError):
                # There are no submodules assiciated with the given commit.
                submodules = []

            subclients = {
                submodule: LocalClient(
                    path=(client.path / submodule.path).resolve(),
                    parent=client,
                )
                for submodule in submodules
            }

            def resolve_submodules(file_, **kwargs):
                original_path = client.path / file_
                if original_path.is_symlink(
                ) or file_.startswith('.renku/vendors'):
                    original_path = original_path.resolve()
                    for submodule, subclient in subclients.items():
                        try:
                            subpath = original_path.relative_to(subclient.path)
                            return Dependency.from_revision(
                                client=subclient,
                                path=str(subpath),
                                revision=submodule.hexsha,
                                submodules=hierarchy + [submodule.name],
                                **kwargs
                            )
                        except ValueError:
                            pass

            for input_id, input_path in process.iter_input_files(basedir):
                try:
                    dependency = resolve_submodules(input_path, id=input_id)
                    if dependency is None:
                        dependency = Dependency.from_revision(
                            client=client,
                            path=input_path,
                            id=input_id,
                            revision=revision,
                        )
                    inputs[input_path] = dependency
                except KeyError:
                    continue

        return cls(
            commit=commit,
            client=client,
            process=process,
            process_path=process_path,
            inputs=inputs,
            outputs=outputs,
            children=children,
            submodules=hierarchy,
        )

    @classmethod
    def build_graph(cls, client, revision='HEAD', lookup=None, graph=None):
        """Build a graph for the whole repository."""
        graph = graph or {}
        if lookup is None:
            lookup = deque(
                Dependency(client=client, commit=commit)
                for commit in client.git.iter_commits(rev=revision)
            )
        else:
            lookup = deque(lookup)

        while lookup:
            dependency = lookup.popleft()
            if dependency.commit in graph:
                continue

            action = graph[dependency.commit] = cls.from_git_commit(
                dependency.commit,
                client=dependency.client,
                submodules=dependency.submodules,
            )
            lookup.extendleft(action.inputs.values())

        return graph

    @classmethod
    def dependencies(cls, client, revision='HEAD', can_be_cwl=False):
        """Build a status."""
        from renku.cli._graph import _safe_path

        if revision == 'HEAD':
            index = client.git.index
        else:
            from git import IndexFile
            index = IndexFile.from_tree(client.git, revision)

        return [
            Dependency.from_revision(
                client=client,
                path=path,
                revision=revision,
            ) for path, _ in index.entries.keys()
            if _safe_path(path, can_be_cwl=can_be_cwl)
        ]



# -*- coding: utf-8 -*-
#
# Copyright 2018 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Graph builder."""

import os
from collections import defaultdict
from functools import lru_cache

import attr

from renku import errors
from renku._compat import Path
from renku.models.cwl.parameter import InputParameter, WorkflowOutputParameter
from renku.models.cwl.types import PATH_TYPES
from renku.models.cwl.workflow import Workflow


def _safe_path(filepath, can_be_cwl=False):
    """Check if the path should be used in output."""
    # Should not be in ignore paths.
    if filepath in {'.gitignore', '.gitattributes'}:
        return False

    # Ignore everything in .renku ...
    if filepath.startswith('.renku'):
        # ... unless it can be a CWL.
        if can_be_cwl and filepath.endswith('.cwl'):
            return True
        return False

    return True


@attr.s(cmp=False)
class Graph(object):
    """Represent the provenance graph."""

    client = attr.ib()
    G = attr.ib()
    commits = attr.ib(default=attr.Factory(dict))

    cwl_prefix = attr.ib(init=False)

    def __attrs_post_init__(self):
        """Derive basic informations."""
        self.cwl_prefix = self.client.cwl_prefix

    @G.default
    def _default_graph(self):
        """Return directional graph."""
        import networkx as nx
        return nx.DiGraph()

    def normalize_path(self, path):
        """Normalize path relative to the Git workdir."""
        start = self.client.path.resolve()
        path = Path(path).resolve()
        return os.path.relpath(str(path), start=str(start))

    def _format_path(self, path):
        """Return a relative path based on the client configuration."""
        return os.path.relpath(str(self.client.path / path))

    def _is_cwl(self, path):
        """Check if the path is a valid CWL file."""
        return path.startswith(self.cwl_prefix) and path.endswith('.cwl')

    @lru_cache(maxsize=1024)
    def find_cwl(self, commit):
        """Return a CWL."""
        cwl = None
        for file_ in commit.stats.files.keys():
            if self.client.is_cwl(file_):
                if cwl is not None:
                    raise ValueError(file_)  # duplicate
                cwl = file_
        return cwl

    def build(
        self, revision='HEAD', paths=None, dependencies=None, can_be_cwl=False
    ):
        """Build graph from paths and/or revision."""
        from renku.models.commit import Action, Dependency

        if paths is None and dependencies is None:
            dependencies = Action.dependencies(
                self.client, revision=revision, can_be_cwl=can_be_cwl
            )
        elif dependencies is None:
            dependencies = [
                Dependency.from_revision(
                    self.client, path=path, revision=revision
                ) for path in paths
            ]

        current_files = {(dependency.commit.hexsha, dependency.path)
                         for dependency in dependencies if dependency.path}
        latest_commits = {path: commit for commit, path in current_files}

        self.commits = Action.build_graph(
            self.client, lookup=dependencies, graph=self.commits
        )

        for action in self.commits.values():
            for key, data in action.iter_nodes():
                path = key[1]
                commit = str(data['commit'])

                if path in latest_commits:
                    latest = latest_commits[path]
                else:
                    latest_dependency = Dependency.from_revision(
                        data['client'],
                        path=data['path'],
                        # revision='{0}'.format(key[0]),
                    )
                    latest = latest_dependency.commit.hexsha
                    # current_files.add((
                    #     latest_dependency.commit.hexsha,
                    #     latest_dependency.path
                    # ))

                if latest and latest != commit:
                    data['latest'] = latest
                    latest_commits.setdefault(path, latest)
                    # current_files.add((latest, path))

                self.G.add_node(key, **data)

            for source, target, data in action.iter_edges():
                self.G.add_edge(source, target, **data)

        return current_files

    @property
    def _output_keys(self):
        """Return a list of the output keys."""
        return [n for n, d in self.G.out_degree() if d == 0]

    def _need_update(self):
        """Yield all files that need to be updated."""
        import networkx as nx
        visited = set()

        for key in nx.topological_sort(self.G):
            assert key not in visited
            need_update = []
            node = self.G.nodes[key]
            latest = node.get('latest')

            if not latest:
                for data in node.get('contraction', {}).values():
                    latest = data.get('latest')
                    if latest:
                        node['latest'] = latest
                        break

            if latest:
                need_update.append(key)

            for parent, _ in self.G.in_edges(key):
                assert parent in visited
                need_update.extend(self.G.nodes[parent]['_need_update'])

            if not latest and need_update:
                need_update.append(key)

            self.G.nodes[key]['_need_update'] = need_update
            visited.add(key)

    def build_status(self, revision='HEAD', can_be_cwl=False):
        """Return files from the revision grouped by their status."""
        status = {
            'up-to-date': {},
            'outdated': {},
            'multiple-versions': {},
            'deleted': {},
        }

        from renku.models.commit import Action
        dependencies = Action.dependencies(
            self.client, revision=revision, can_be_cwl=can_be_cwl
        )
        current_files = set(self.build(dependencies=dependencies))

        # Prepare status info for each file.
        self._need_update()

        # First find all up-to-date nodes.
        up_to_date = {
            filepath: commit
            for (commit,
                 filepath), need_update in self.G.nodes.data('_need_update')
            if not need_update
        }

        for commit, filepath in current_files:
            if filepath in up_to_date:  # trick the workflow step
                # FIXME use the latest commit
                status['up-to-date'][filepath] = up_to_date[filepath]
            else:
                try:
                    need_update = self.G.nodes[(commit,
                                                filepath)]['_need_update']
                    status['outdated'][filepath] = [need_update]
                except KeyError:
                    pass

        # Merge all versions of used inputs in outdated file.
        multiple_versions = defaultdict(set)

        for need_updates in status['outdated'].values():
            for need_update in need_updates:
                for commit, filepath in need_update:
                    multiple_versions[filepath].add((commit, filepath))

        for commit, filepath in current_files:
            if filepath in multiple_versions:
                multiple_versions[filepath].add((commit, filepath))

        status['multiple-versions'] = {
            key: value
            for key, value in multiple_versions.items() if len(value) > 1
        }

        # Build a list of used files that have been deleted.
        current_paths = {filepath for _, filepath in current_files}
        status['deleted'] = {
            filepath: (commit, filepath)
            for commit, filepath in self.G.nodes
            if _safe_path(filepath, can_be_cwl=can_be_cwl) and filepath not in
            current_paths and not ((self.client.path / filepath).exists() or
                                   (self.client.path / filepath).is_dir())
        }
        return status

    @property
    def output_files(self):
        """Return a list of nodes representing output files."""
        for key in self._output_keys:
            node = self.G.nodes[key]
            if 'tool' not in node:
                yield key, node

    @property
    def _tool_nodes(self):
        """Yield topologically sorted tools."""
        import networkx as nx

        for key in nx.topological_sort(self.G):
            node = self.G.nodes[key]
            tool = node.get('tool')
            if tool is not None:
                yield key, node

    def siblings(self, key):
        """Return siblings for a given key.

        The key is part of the result set, hence to check if the node has
        siblings you should check the lenght is greater than 1.
        """
        parents = list(self.G.predecessors(key))
        if not parents:
            raise errors.InvalidOutputPath(
                'The file "{0}" was not created by a renku command. \n\n'
                'Check the file history using: git log --follow "{0}"'.format(
                    key[1]
                )
            )
        return {
            sibling
            for parent in parents for sibling in self.G.successors(parent)
        }

    def ascwl(self, global_step_outputs=False):
        """Serialize graph to CWL workflow.

        :param global_step_outputs: Make all step outputs global.
        """
        workflow = Workflow()

        input_index = 1
        steps = {}

        def _source_name(key):
            """Find source name for a node."""
            if self.G.in_degree(key) == 0:
                return None

            assert self.G.in_degree(key) == 1

            tool_key, node = list(self.G.pred[key].items())[0]
            return '{0}/{1}'.format(steps[tool_key], node['id'])

        def _relative_default(client, default):
            """Evolve ``File`` or ``Directory`` path."""
            if isinstance(default, PATH_TYPES):
                path = (client.workflow_path / default.path).resolve()
                return attr.evolve(default, path=path)
            return default

        for tool_index, (key, node) in enumerate(self._tool_nodes, 1):
            _, path = key
            tool = node['tool']
            step_id = 'step_{0}'.format(tool_index)
            steps[key] = step_id

            ins = {
                edge_id: _source_name(target_id)
                for target_id, _, edge_id in self.G.in_edges(key, data='id')
            }
            outs = [
                edge_id for _, _, edge_id in self.G.out_edges(key, data='id')
            ]

            for input_ in tool.inputs:
                input_mapping = ins.get(input_.id)
                if input_mapping is None:
                    input_id = 'input_{0}'.format(input_index)
                    workflow.inputs.append(
                        InputParameter(
                            id=input_id,
                            type=input_.type,
                            default=_relative_default(
                                self.client, input_.default
                            ),
                        )
                    )
                    input_index += 1
                    ins[input_.id] = input_id

            workflow.add_step(
                run=self.client.path / path,
                id=step_id,
                in_=ins,
                out=outs,
            )

        if global_step_outputs:
            output_keys = (key for _, key in self.G.out_edges(steps.keys()))
        else:
            output_keys = self._output_keys

        for index, key in enumerate(output_keys):
            output_id = 'output_{0}'.format(index)
            # FIXME use the type of step output
            type_ = 'Directory' if (self.client.path /
                                    key[0]).is_dir() else 'File'
            workflow.outputs.append(
                WorkflowOutputParameter(
                    id=output_id,
                    type=type_,
                    outputSource=_source_name(key),
                )
            )

        return workflow
